import torch
import matplotlib.pyplot as plt
from matplotlib.widgets import Cursor
from lxfx.data import TimeSeriesDatasetManager

class PredPlotter():
    """
    This handles all instances of plotting predictions in different ways.
    """
    def __init__(self, preds, dataset_manager:TimeSeriesDatasetManager,
                 autoreg_batch_plots:bool = False, test_only:bool = False, 
                 length:int = None, show_gridlines = True, interactive_cursor = True,
                 figsize = (6,6), test_plots = 1, is_production = False):
        self.is_production = is_production
        self.preds = preds
        self.dataset_manager = dataset_manager
        self.autoreg_batch_plots = autoreg_batch_plots
        self.test_plots = test_plots
        self.test_only = test_only
        self.length = length
        self.show_gridlines = show_gridlines
        self.interactive_cursor = interactive_cursor

        self.figsize = figsize
        self.figure = plt.figure(figsize = self.figsize)
        self.num_rows = self.test_plots
        self.idx_count = 0

        self.train_targets = self.dataset_manager.train_targets
        self.val_targets = self.dataset_manager.val_targets
        self.test_data = self.dataset_manager.test_targets
        self.test_eq_target_seqs = self.dataset_manager.test_eq_target_seqs
        self.test_target_seqs = self.dataset_manager.test_seq_targets
        self.transform = self.dataset_manager.transform if self.dataset_manager.transform is not None else self.dataset_manager.default_transform

        self.plot_train_targets = None
        self.plot_val_targets = None
        self.plot_test_data = None
        self.plot_pred_data = None

        self.train_y_axes = None 
        self.val_y_axes = None 
        self.test_y_axes = None 
        self.pred_y_axes = None

        self.train_x_axes = None
        self.val_x_axes = None
        self.test_x_axes = None
        self.pred_x_axes = None

        self.val_start_index = None
        self.val_end_index = None
        self.test_start_index = None
        self.test_end_index = None
        self.pred_start_index = None
        self.pred_end_index = None

    def get_prev_batch_time_steps(self, seq_index = None):
        if seq_index > 0:
            additional_seqs = self.test_eq_target_seqs[:seq_index]
        else:
            additional_seqs = []
        return additional_seqs
    
    def get_plot_data(self, preds, seq_index = None, column_idx = 0):
        self.plot_train_targets = self.dataset_manager.inverse_targets(self.train_targets)[:,column_idx]

        self.plot_val_targets = self.dataset_manager.inverse_targets(self.val_targets)[:,column_idx]

        if self.autoreg_batch_plots:
            # add previous time steps from the test data to the plot_val_targets
            additional_seqs = self.get_prev_batch_time_steps(seq_index)
            for seq in additional_seqs:
                seq = self.dataset_manager.inverse_targets(seq)[:,column_idx]
                self.plot_val_targets = torch.cat((self.plot_val_targets, seq), dim=0)

            plot_test_data = torch.cat((self.test_eq_target_seqs[seq_index], self.test_target_seqs[seq_index]), dim=0)
            self.plot_test_data = self.dataset_manager.inverse_targets(plot_test_data)[:,column_idx]

        else:
            self.plot_test_data = self.dataset_manager.inverse_targets(self.test_data)[:,column_idx]
        # preds for the column of column index column_idx are the ones provided in the preds parameter
        self.plot_pred_data = self.dataset_manager.inverse_targets(preds)

    def get_y_axes(self):
        self.train_y_axes = self.plot_train_targets 
        self.val_y_axes = self.plot_val_targets

        # length only affects the test data and predictions when not autoregressive
        if not self.autoreg_batch_plots:    
            if self.length is None:
                self.test_y_axes = self.plot_test_data
                self.pred_y_axes = self.plot_pred_data
            else:
                if self.length > len(self.plot_test_data):
                    self.length = len(self.plot_test_data)
                self.test_y_axes = self.plot_test_data[:self.length]
                self.pred_y_axes = self.plot_pred_data[self.dataset_manager.sequenceLength:self.length]
        else:
            # We want to see the whole autoregressive prediction vs the test curve
            self.test_y_axes = self.plot_test_data
            self.pred_y_axes = self.plot_pred_data

    def get_plot_indices(self, seq_idx = 0):
        self.val_start_index = len(self.train_y_axes)
        self.val_end_index = len(self.train_y_axes)+len(self.val_y_axes)
        if self.autoreg_batch_plots:
            self.test_start_index = self.val_end_index
            if self.is_production:
                self.test_end_index = self.val_end_index+len(self.test_eq_target_seqs[seq_idx])
                self.pred_start_index = self.test_end_index
                self.pred_end_index = self.pred_start_index+self.preds[seq_idx].shape[1] # We add on the length of the future predicted points
            else:
                self.test_end_index = self.test_start_index+len(self.test_eq_target_seqs[seq_idx])+self.preds[seq_idx].shape[1]
                self.pred_start_index = self.val_end_index+len(self.test_eq_target_seqs[seq_idx])
                self.pred_end_index = self.test_end_index
            # Here the test_data is each test_seq in the test_seqs
        else:
            self.test_start_index = self.val_end_index
            self.test_end_index = self.val_end_index+len(self.test_y_axes)
            self.pred_start_index = self.test_start_index + self.dataset_manager.sequenceLength

    def get_x_axes(self):
        self.train_x_axes = range(len(self.train_y_axes))
        self.val_x_axes = range(self.val_start_index, self.val_end_index)
        if not self.test_only:
            self.test_x_axes = range(self.val_end_index, self.test_end_index)
            self.pred_x_axes = range(self.pred_start_index, self.test_end_index)
        else:
            self.test_x_axes = range(self.test_start_index, self.test_end_index)
            self.pred_x_axes = range(self.pred_start_index, self.test_end_index)

    def plotColumnPreds(self, preds:torch.Tensor, seq_index:int = None,
                        column_idx:int = 0, seq_idx = 0):
        self.get_plot_data(preds, seq_index, column_idx)
        self.get_y_axes()
        self.get_plot_indices(seq_idx)
        self.get_x_axes()

        if isinstance(self.preds, torch.Tensor):
            num_columns = self.preds.shape[-1]
        else:
            num_columns = self.preds[seq_idx].shape[-1]
        self.idx_count += 1
        # Plotting
        # fig = plt.figure(figsize = figsize)
        axs = self.figure.add_subplot(self.num_rows, num_columns, self.idx_count)
        # axs.set_title(f"Predictions for column {self.dataset_manager.df.columns[column_idx]}")
        
        if self.autoreg_batch_plots:
            test_x_axes = self.test_x_axes[:(len(self.test_x_axes)-len(self.pred_x_axes))]
            test_labels_x_axes = self.test_x_axes[-(len(self.pred_x_axes)):]

            test_y_axes = self.test_y_axes[:(len(self.test_x_axes)-len(self.pred_x_axes))]
            test_labels_y_axes = self.test_y_axes[-(len(self.pred_x_axes)):]
        else:
            test_x_axes = self.test_x_axes
            test_y_axes = self.test_y_axes
            test_labels_x_axes = None 
            test_labels_y_axes = None

        if not self.test_only:
            axs.plot(self.train_x_axes, self.train_y_axes, c = "black", label = "train")
            axs.plot(self.val_x_axes, self.val_y_axes, c = "green", label = "val")
        axs.plot(test_x_axes, test_y_axes, c = "yellow", label = "test")
        if self.autoreg_batch_plots:
            axs.plot(test_labels_x_axes, test_labels_y_axes, c = "green", label = "true_preds")
        axs.plot(self.pred_x_axes, self.pred_y_axes, c = "red", label = "preds")
        axs.set_title(f"{self.dataset_manager.target_column_names[column_idx]}")
        axs.legend()

        if self.interactive_cursor:
            cursor = Cursor(axs, useblit=True, color='red', linewidth=1)
        if self.show_gridlines:
            axs.grid(True)

    def ColumnsPlot(self, future_preds):
        main_fig = plt.figure(figsize=self.figsize)
        figs = []
        for t in range(future_preds.shape[-1]):
            feature_future_preds = future_preds[:,t]
            figs.append(self.plotColumnPreds(feature_future_preds))

        # for index, fig in enumerate(figs):
        #     main_fig.add_subfigure(fig) # More to be done here


    def plotAutoRegPreds(self):
        if isinstance(self.preds, list): # We have an autogressive prediction model

            # handle each autoregressive prediction
            for seq_idx, future_preds in enumerate(self.preds):
                if seq_idx >= self.test_plots:
                    break
                # Note: batch dim is at index 0
                future_preds = future_preds.squeeze(0) # remove the batch dimension

                # handle multiple columns
                if future_preds.shape[-1] > 1:
                    for t in range(future_preds.shape[-1]):
                        feature_future_preds = future_preds[:,t].unsqueeze(1)

                        self.plotColumnPreds(feature_future_preds, seq_index = seq_idx,
                                             column_idx = t, seq_idx = seq_idx)
                else:
                    self.plotColumnPreds(future_preds, seq_index=seq_idx)
            plt.show()

    def plotTimeStepPreds(self):
        future_preds = self.preds
        # handle multiple columns
        if future_preds.shape[-1] > 1:
            future_preds = future_preds.squeeze(1)
            for t in range(future_preds.shape[-1]):
                feature_future_preds = future_preds[:,t].unsqueeze(1)
                self.plotColumnPreds(feature_future_preds, column_idx=t)
        else:
            self.plotColumnPreds(future_preds)
        plt.show()

    def plotPredictions(self):
        if isinstance(self.preds, list):
            self.autoreg_batch_plots = True
            self.plotAutoRegPreds()
        else:
            self.autoreg_batch_plots = False
            self.plotTimeStepPreds()


import torch
import numpy as np
import pandas as pd
from torch.utils.data import Dataset, DataLoader
from typing import Union
import logging
from pathlib import Path
import matplotlib.pyplot as plt
from statsmodels.tsa import seasonal
import plotly.graph_objects as graph_objects
import plotly.subplots as subplots

from lxfx.utils import createLogger, handleExit
from lxfx.metrics import TorchMinMaxScaler, TorchStandardScaler

class TimeSeriesDatasetManager():
    class TimeSeriesDataset(Dataset):
        def __init__(self, data:torch.Tensor, targets:torch.Tensor):
            super().__init__()
            self.data = data 
            self.targets= targets 
            # self.test = test
            # self.batch_size = batch_size

        def __len__(self):
            return len(self.data)
        
        def __getitem__(self, index):
            # if index < s
            return self.data[index], self.targets[index]
        
    def __init__(self, data: Union[torch.Tensor, np.ndarray, pd.DataFrame] = None,
                file_path = None,
                targets:torch.Tensor|np.ndarray = None,
                future_pred_length = 1,
                targetIndices:list = None,
                target_names:list = None,
                drop_names:list = None,
                date_col:str = None,
                sequence_length = 5,
                stride = 1,
                scaler_feature_range = (0,1),
                transform = None,
                use_default_scaler = True,
                split_criterion = (0.7, 0.15, 0.15),
                batchsize = None, 
                is_fx = False,
                transform_targets = True,
                manual_seed = None,
                is_autoregressive = False,
                test_sequence_length = 2,
                test_stride = None,
                **kwargs):
        """
        Initializes the TimeSeriesDatasetManager with the provided data and parameters.

        Parameters:
            data (torch.Tensor|np.ndarray|pd.DataFrame, optional): The input data for the time series. Defaults to None.
            file_path (str, optional): The file path to load the data from. Defaults to None.
            targets (torch.Tensor|np.ndarray, optional): The target values for the time series. Defaults to None.
            future_pred_length (int, optional): The length of the future predictions. Defaults to 1.
            targetIndices (list, optional): The indices of the target columns. Defaults to None.
            target_names (list, optional): The names of the target columns. Defaults to None.
            drop_names (list, optional): The names of the columns to drop. Defaults to None.
            date_col (str, optional): The name of the date column. Defaults to None.
            sequence_length (int, optional): The length of the sequences to create. Defaults to 5.
            stride (int, optional): The stride between sequences. Defaults to 1.
            scaler_feature_range (tuple, optional): The feature range for the scaler. Defaults to (0, 1).
            transform (callable, optional): The transform to apply to the data. Defaults to None.
            use_default_scaler (bool, optional): Whether to use the default scaler. Defaults to True.
            split_criterion (tuple, optional): The split criterion for train, validation, and test sets. Defaults to (0.7, 0.15, 0.15).
            batchsize (int, optional): The batch size for training. Defaults to None.
            is_fx (bool, optional): if true then log returns are to be calculated with a default multiplier of 10000
            transform_targets: Whether or not to transform the targets using the transform or scaler used to transform the data
            manual_seed (int, optional): The manual seed to set for the random number generator. Defaults to None.
            is_autoregressive (bool, optional): Whether the model is autoregressive. Defaults to False.
            test_sequence_length (int, optional): The length of the test sequences to create. Defaults to 2.
            **kwargs: Additional keyword arguments.

        """
        # todo: cater for the data if available instead of just dropping it
        super().__init__()
        self.console_logger = createLogger(is_consoleLogger=True, log_level = logging.WARNING)

        self.manual_seed = manual_seed
        if manual_seed:
            self.set_seed(self.manual_seed)

        self.data = data 
        self.is_fx = is_fx
        self.targets = targets 
        self.sequenceLength = sequence_length
        self.is_autoregressive = is_autoregressive
        self.test_sequence_length = test_sequence_length
        self.future_pred_length = future_pred_length
        self.data_filepath = file_path
        self.targetIndices = targetIndices
        self.targetNames = target_names
        self.kwargs = kwargs
        self.drop_names = drop_names
        self.date_col_name = date_col
        self.data_column_names = []
        self.target_column_names = []
        self.date_series = None
        self.batchsize = batchsize
        self.stride = stride
        self.test_stride = test_stride
        self.target_dict = {}
        self.use_default_scaler = use_default_scaler
        self.is_data_transformed = False 
        self.is_targets_transformed = False 
        self.scaler_feature_range = scaler_feature_range
        self.transform = transform
        if self.transform:
            if self.transform == "z-norm":
                self.transform = TorchStandardScaler()
        self.default_transform = TorchMinMaxScaler(feature_range = self.scaler_feature_range)
        self.df_column_names = []
        self.df = None

        self.getData() # get the data ready

        self.dataset_size = len(self.data)

        self.test_batchsize = 1
        self.train_split = split_criterion[0]
        self.val_split = split_criterion[1]
        self.test_split = split_criterion[2]

        self.train_data = None 
        self.test_data = None 
        self.val_data = None
        self.train_targets = None 
        self.test_targets = None
        self.val_targets = None  
        self.val_end_index = None 
        self.test_start_index = None
        self.transform_targets = transform_targets
        # split the data
        self.splitData()

        self.train_data_seqs = None 
        self.train_eq_target_seqs = None 
        self.test_data_seqs = None 
        self.test_eq_target_seqs = None 
        self.val_data_seqs = None 
        self.val_eq_target_seqs = None 
        self.train_seq_targets = None 
        self.val_seq_targets = None 
        self.test_seq_targets = None
        self.createSequences()

        self.train_dataloader = None
        self.test_dataloader = None 
        self.val_dataloader = None 

        self.generateDataloaders()

    def config_dict(self):
        """
        Returns a dictionary representation of the instance's attributes.
        """
        return {
            "file_path": self.data_filepath,
            "targetIndices": self.targetIndices,
            "target_names": self.targetNames,
            "drop_names": self.drop_names,
            "date_col": self.date_col_name,
            "sequence_length": self.sequenceLength,
            "stride": self.stride,
            "scaler_feature_range": self.scaler_feature_range,
            "transform": self.transform,
            "use_default_scaler": self.use_default_scaler,
            "split_criterion": (self.train_split, self.val_split, self.test_split),
            "batchsize": self.batchsize,
            "is_fx": self.is_fx,
            "manual_seed": self.manual_seed,
            "transform_targets": self.transform_targets,
        }
    def printSummary(self, train_only = False, val_only = False,
                     test_only = False, sample_size = 15,
                     seq_sample_size = 3):
        """
        print a summary of what has been done to the data
        """
        summary = {}
        summary["Data loaded from"] = self.data_filepath
        summary["Data columns"] = self.df_column_names
        summary["Data shape"] = self.data.shape
        summary["Targets shape"] = self.targets.shape

        summary["split criterion"] = (self.train_split, self.val_split, self.test_split)
        summary["train_data_shape"] = self.train_data.shape
        summary["val_data_shape"] = self.val_data.shape
        summary["test_data_shape"] = self.test_data.shape
        summary["train_targets_shape"] = self.train_targets.shape
        summary["val_targets_shape"] = self.val_targets.shape
        summary["test_targets_shape"] = self.test_targets.shape

        summary["sequence_length"] = self.sequenceLength
        summary["future_pred_length"] = self.future_pred_length
        summary["stride"] = self.stride

        for k, v in summary.items():
            print(f"{k}: {v}")

        print(f"\nData sample_size: {sample_size}")
        print(f"Sequence length: {self.sequenceLength}")
        print(f"Sequence sample_size: {seq_sample_size}")
        
        if train_only:
            print(f"Data : {self.data[:sample_size]}")
            print(f"Train data: {self.train_data[:sample_size]}\n")
            print(f"Train targets: {self.train_targets[:sample_size]}\n")

        if val_only:
            val_start_index = len(self.train_data)
            print(f"Data from val_start_index: {self.data[val_start_index:(val_start_index+5)]}")
            print(f"Val data: {self.val_data[:sample_size]}\n")
            print(f"Val targets: {self.val_targets[:sample_size]}\n")

        if test_only:
            print(f"Data from test_start_index: {self.data[self.test_start_index:(self.test_start_index+15)]}")
            print(f"Test data: {self.test_data[:sample_size]}\n")
            print(f"Test targets: {self.test_targets[:sample_size]}\n")

        print("\tSequences")
        if train_only:
            print(f"Train sequences: {self.train_data_seqs[:seq_sample_size]}")
            print(f"Train sequence targets: {self.train_seq_targets[:seq_sample_size]}\n")
        if val_only:
            print(f"Val sequences: {self.val_data_seqs[:seq_sample_size]}")
            print(f"Val sequence targets: {self.val_seq_targets[:seq_sample_size]}\n")
        if test_only:
            print(f"Test sequences: {self.test_data_seqs[:seq_sample_size]}")
            print(f"Test sequence targets: {self.test_seq_targets[:seq_sample_size]}\n")

        print("\tBatches")
        if train_only:
            print(f"Train batchsize: {self.train_dataloader.batch_size}\n")
        if val_only:
            print(f"Val batchsize: {self.val_dataloader.batch_size}\n")
        if test_only:
            print(f"Test batchsize: {self.test_dataloader.batch_size}\n")

    def processDataFrame(self, df:pd.DataFrame):
        if "index_col" in self.kwargs.keys():
            df = df.reset_index(drop = True) # remove the index column
        if self.date_col_name:
            df[self.date_col_name] = pd.to_datetime(df[self.date_col_name])
            self.date_series = df[self.date_col_name].copy()
        # remove the target columns such that the model does not have to see them.
        # first save them as the target labels 
        if self.targetNames:
            for tN in self.targetNames:
                self.target_dict[tN] = df[tN]
        elif self.targetIndices:
            for tI in self.targetIndices:
                self.target_dict[tI] = df[df.columns[tI]]

        # remove the target columns from the data frame
        if not (self.date_col_name and (len(self.df.columns) == 2)): # there is only a single row of data which is both the target and the feature
            if self.targetIndices:
                [df.drop(columns = df.columns[tI]) for tI in self.targetIndices]
            elif self.targetNames: 
                for tN in self.targetNames:
                    df = df.drop(columns = tN)
            else:
                self.console_logger.error("Neither Target index nor targetNames provided")
                handleExit()
        else:
            self.console_logger.info("Dataframe has only a single feature which is at the same time the target")
        
        # remove any other columns not wanted forexample Date as provided by the user in the drop_names list
        try:
            if self.drop_names:
                for drop_name in self.drop_names: 
                    df = df.drop(columns = drop_name)
        except Exception as e:
            self.console_logger.error(f"Error dropping columns: {e}")
            handleExit()
        # now cater for the targets 
        target_df = pd.DataFrame(self.target_dict)
        # round off the values in the dataframe to 5dps
        df = df.round(5)
        self.df_column_names.extend(df.columns.to_list())
        target_df = target_df.round(5)
        # set the class variables appropriately 
        self.data_column_names = df.columns.copy()
        self.target_column_names = target_df.columns.copy()
        self.data = torch.tensor(df.to_numpy(dtype=np.float32), dtype = torch.float32)
        self.targets = torch.tensor(target_df.to_numpy(dtype=np.float32), dtype = torch.float32)

    def fetchData(self):
        if self.data_filepath:
            if Path(self.data_filepath).is_file():
                try:
                    index_col = None
                    if "index_col" in self.kwargs.keys():
                        index_col = self.kwargs["index_col"]
                    df = pd.read_csv(self.data_filepath,index_col=index_col)
                    self.df = df.copy()
                except Exception as e:
                    self.console_logger.error(f"Error reading file: {e}")
                    handleExit()
                self.processDataFrame(df)

            else:
                __log = f"File not Found: {self.data_filepath}"
                self.console_logger.error(__log)
                # file_logger.error(__log)
        if isinstance(self.data, pd.DataFrame):
            self.df = self.data.copy()
            self.processDataFrame(self.data)

    def getData(self):
        if self.data_filepath:
            self.fetchData()
        if isinstance(self.data, torch.Tensor):
            pass
        elif isinstance(self.data, pd.DataFrame):
            self.fetchData()

        # Note: by this step all the data is of type np.ndarray

        # calculate the log returns before rescaling the values
        if self.is_fx:
            # Calculate log returns for the first three columns and retain the volume column
            if "Volume" in self.drop_names:
                self.data = self.calculateLogReturns(prices=self.data, fx=True)
                self.targets = self.calculateLogReturns(prices=self.targets, fx=True)                
            else:
                log_returns = self.calculateLogReturns(prices=self.data[:, :3], fx=True)
                self.data = torch.column_stack((log_returns, self.data[1:, 3])) 
                self.targets = self.calculateLogReturns(prices=self.targets, fx=True)

    def generateSequences(self, train = False, test= False, val=False):
        data_seqs = []
        eq_target_seqs = []
        target_seqs = []

        sequence_length = self.sequenceLength
        stride =self.stride
        if train:
            data = self.train_data 
            targets = self.train_targets 
        elif test: 
            if self.is_autoregressive:
                if self.test_stride:
                    stride = self.test_stride
            data = self.test_data 
            targets = self.test_targets 
            if self.is_autoregressive:
                sequence_length = self.test_sequence_length
        elif val: 
            data = self.val_data 
            targets = self.val_targets

        list_data = data.tolist()
        list_targets = targets.tolist()

        for idx in range(0, len(list_data)-sequence_length, stride):
            if idx + sequence_length + self.future_pred_length <= len(list_targets):
                data_seq = list_data[idx:idx+sequence_length]
                eq_target_seq = list_targets[idx:idx+sequence_length]
                target = list_targets[idx+sequence_length:idx+sequence_length+self.future_pred_length]
                data_seqs.append(data_seq)
                eq_target_seqs.append(eq_target_seq)
                target_seqs.append(target) if isinstance(target, list) else target_seqs.append([target])

        # transform the lists into tensors
        data_tensor_seqs = torch.tensor(data_seqs, dtype=torch.float32)
        eq_target_tensor_seqs = torch.tensor(eq_target_seqs, dtype=torch.float32)
        target_tensor_seqs = torch.tensor(target_seqs, dtype=torch.float32)
        return data_tensor_seqs, target_tensor_seqs, eq_target_tensor_seqs

    def showSeriesComponents(self, period, model = "additive", column_name = None,
                             column_idx=None, height = 800):
        """
        Decomposes and visualizes the components of a time series.

        This function decomposes a time series into its observed, trend, seasonal, and residual components
        using seasonal decomposition. It then visualizes these components using a plot with four subplots.

        Parameters:
            period (int): The period for seasonal decomposition.
            model (str): The type of seasonal decomposition model to use. Default is "additive".
            column_name (str): The name of the column in the DataFrame to decompose.
            column_idx (int): The index of the column in the DataFrame to decompose. Default is None.
            height (int): The height of the plot. Default is 800.

        Returns:
        None
        """
        titles = ["observed", "trendComponent", "seasonalComponent", "residualComponent"]
        colors = ["blue", "green", "orange", "red"]
        data = self.df[column_name]
        dc = seasonal.seasonal_decompose(
            data,
            model = model, 
            period = period
        )
        x_axes = self.df.index
        componets = [dc.observed, dc.trend, dc.seasonal, dc.resid]
        fig = subplots.make_subplots(4, 1, shared_xaxes=True, subplot_titles=titles)
        for idx, (cmp, title, color) in enumerate(zip(componets, titles, colors)):
            scatter_trace= graph_objects.Scatter(x = x_axes, y = cmp,
                                                mode="lines", name=title,
                                                line=dict(color=color, width=2))
            fig.add_trace(scatter_trace,row=idx+1,col=1)
        fig.update_layout(height=height, title=f'Decomposed Analysis of {column_name}',    
                  xaxis_title='Date', yaxis_title='Value', showlegend=True)
        fig.show()

    def createSequences(self):
        """
        Create sequences for each of the datasets ie train_data, val_data, test_data
        """
        # generate train_sequences 
        self.train_data_seqs, self.train_seq_targets, self.train_eq_target_seqs = self.generateSequences(train= True)
        self.test_data_seqs, self.test_seq_targets, self.test_eq_target_seqs = self.generateSequences(test= True)
        self.val_data_seqs, self.val_seq_targets, self.val_eq_target_seqs = self.generateSequences(val= True)

    def transformTargets(self, transform):
        if self.is_targets_transformed:
            self.console_logger.info("Targets already transformed. Not transforming")
        else:
            if self.transform_targets:
                # transform the targets using the same transform that was used on the train_data repeat 
                if self.train_data.shape[-1] > 1:
                    num_features = self.train_targets.shape[-1]
                    # Note: tile is used to repeat the targets the number of times as the number of features in the train_data since the transform was fitted on the train_data 
                    if isinstance(self.train_targets, np.ndarray):
                        train_targets_placeholder = np.zeros((self.train_targets.shape[0], 1))
                        val_targets_placeholder = np.zeros((self.val_targets.shape[0], 1))
                        test_targets_placeholder = np.zeros((self.test_targets.shape[0], 1))
                        # concatenate the targets with the placeholder for a specific number of times
                        for i in range(self.train_data.shape[-1] - self.train_targets.shape[-1]):
                            self.train_targets = np.concatenate((self.train_targets, train_targets_placeholder), axis=1)
                            self.val_targets = np.concatenate((self.val_targets, val_targets_placeholder), axis=1)
                            self.test_targets = np.concatenate((self.test_targets, test_targets_placeholder), axis=1)

                        self.train_targets = transform.transform(self.train_targets)[:, 0:(num_features)]
                        self.val_targets = transform.transform(self.val_targets)[:, 0:(num_features)]
                        self.test_targets = transform.transform(self.test_targets)[:, 0:(num_features)]

                    elif isinstance(self.train_targets, torch.Tensor):
                        train_targets_placeholder = torch.zeros(self.train_targets.shape[0], 1)
                        val_targets_placeholder = torch.zeros(self.val_targets.shape[0], 1) 
                        test_targets_placeholder = torch.zeros(self.test_targets.shape[0], 1)
                        # concatenate the targets with the placeholder for a specific number of times
                        for i in range(self.train_data.shape[-1]-self.train_targets.shape[-1]):
                            self.train_targets = torch.cat((self.train_targets, train_targets_placeholder), dim=1)
                            self.val_targets = torch.cat((self.val_targets, val_targets_placeholder), dim=1)
                            self.test_targets = torch.cat((self.test_targets, test_targets_placeholder), dim=1)

                        self.train_targets = transform.transform(self.train_targets)[:,0:(num_features)]
                        self.val_targets = transform.transform(self.val_targets)[:,0:(num_features)]
                        self.test_targets = transform.transform(self.test_targets)[:,0:(num_features)]
                else:
                    self.train_targets = transform.transform(self.train_targets)
                    self.val_targets = transform.transform(self.val_targets)
                    self.test_targets = transform.transform(self.test_targets)

    def transform_data(self):
        # rescale the data

        if self.is_data_transformed == False:
            if self.transform is None:
                if self.use_default_scaler:
                    # fit the transform on the train data and use it to transform the other data
                    self.default_transform.fit(self.train_data)
                    # use the fitted transform to transform even the val and test data
                    self.train_data = self.default_transform.transform(self.train_data)
                    self.val_data = self.default_transform.transform(self.val_data)
                    self.test_data = self.default_transform.transform(self.test_data)

                    self.transformTargets(self.default_transform)

            else:
                    # fit the transform on the train data and use it to transform the other data
                    self.transform = self.transform.fit(self.train_data)
                    # use the fitted transform to transform even the val and test data
                    self.train_data = self.transform.transform(self.train_data)
                    self.val_data = self.transform.transform(self.val_data)
                    self.test_data = self.transform.transform(self.test_data)

                    self.transformTargets(self.transform)

            self.is_data_transfomed = True 
        else:
            self.console_logger.info("Data already transformed. Not transforming")

    def splitData(self):
        # note that slicing does not create a copy of the original data it just creates a view
        train_size = int(self.train_split*self.dataset_size)
        val_size = int(self.val_split*self.dataset_size)
        test_size = self.dataset_size - (train_size + val_size)

        self.val_end_index = train_size 
        self.test_start_index = (train_size+val_size)

        self.train_data = self.data[:self.val_end_index]
        self.train_targets = self.targets[:self.val_end_index]

        if val_size != 0:
            self.val_data = self.data[self.val_end_index:self.test_start_index]
            self.val_targets = self.targets[self.val_end_index:self.test_start_index]
        if test_size != 0:
            self.test_data = self.data[self.test_start_index:-(self.batchsize)]
            self.test_targets = self.targets[self.test_start_index:-(self.batchsize)]

        self.transform_data()

    def datasets(self):
        """ 
        Returns:
            train_dataset, val_dataset, test_dataset (all of which are of type 
            torch.utils.data.Dataset)
        """
        self.splitData()
        train_dataset = self.TimeSeriesDataset(self.train_data_seqs, self.train_seq_targets)
        val_dataset = self.TimeSeriesDataset(self.val_data_seqs, self.val_seq_targets)
        test_dataset = self.TimeSeriesDataset(self.test_data_seqs, self.test_seq_targets)
        return train_dataset, val_dataset, test_dataset

    def generateDataloaders(self):
        train_dataset, val_dataset, test_dataset =  self.datasets()
        self.train_dataloader = DataLoader(train_dataset, shuffle = False, batch_size = self.batchsize)
        self.val_dataloader =  DataLoader(val_dataset, shuffle = False, batch_size = self.batchsize)
        self.test_dataloader = DataLoader(test_dataset, shuffle = False, batch_size = self.test_batchsize)

    def dataloaders(self):
        """
        Returns:
            train_dataloader, val_dataloader test_dataloader (all of which are of type
            torch.utils.data.DataLoader)
        """
        return self.train_dataloader, self.val_dataloader, self.test_dataloader
    
    def plotProportionGraph(self, figsize = (10, 10), column_name = None):
        """
        Plots a graph showing the proportions of the train, validation, and test data.
        """
        if (len(self.df.columns) ==2) and self.date_col_name : # we have only a single data column which is at the same time the target
            train_axes = self.train_data
            val_axes = self.val_data
            test_axes = self.test_data
            val_start_index = len(train_axes)
            val_end_index = len(train_axes)+len(val_axes)
            test_end_index = val_end_index+len(test_axes)
        else:
            train_axes = self.train_data[:,1]
            val_axes = self.val_data[:,1]
            test_axes = self.test_data[:,1]
            val_start_index = len(train_axes)
            val_end_index = len(train_axes)+len(val_axes)
            test_end_index = val_end_index+len(test_axes)
        fig  = plt.figure(figsize = figsize)
        axs = fig.add_subplot(1,1,1)
        axs.set_title("Graph showing test val test proportions")
        axs.plot(range(len(train_axes)), train_axes, c = "black", label = "train")
        axs.plot(range(val_start_index, val_end_index), val_axes, c = "green", label = "validation")
        axs.plot(range(val_end_index, test_end_index), test_axes, c = "yellow", label = "test")
        fig.legend()

    def plotSampleColumn(self, column_name,figsize=(6,6)):
        """
        Plots a graph showing a sample of a column from the data.

        Parameters:
            column_name (str): The name of the column to plot.
            figsize (tuple, optional): The size of the figure. Defaults to (6, 6).

        Returns:
            None
        """
        column_name_idx = None
        for idx, cN in enumerate(self.df_column_names):
            if cN == column_name:
                column_name_idx = idx
                break
        if column_name_idx:
            fig = plt.figure(figsize = figsize)
            axs = fig.add_subplot(1,1,1)
            data = self.data[:,column_name_idx]
            axs.plot(data)
        else:
            self.console_logger.error(f"Column name: {column_name} not found")

    def convertLogReturnsToCurPrices(self, prevPrcs = None, log_returns = None, fx=True):
        """
        Converts the log returns to the current prices.

        Parameters:
            prevPrcs: The previous prices
            log_returns: The log returns
            fx: Whether the data is for foreign exchange
        """
        if isinstance(log_returns, np.ndarray):
            log_returns = torch.from_numpy(log_returns)
        if isinstance(prevPrcs, np.ndarray):
            prevPrcs = torch.from_numpy(prevPrcs)
        prices = torch.exp(log_returns) * prevPrcs
        if fx:
            prices = prices*10000
        return prices

    def calculateLogReturns(self, curPrc = None, prevPrc = None, prices = None, fx=True, cum = None):
        """
        Calculates the log returns of a time series.

        Parameters:
            curPrc: The current price
            prevPrc: The previous price
            prices: The prices of the time series
            fx: Whether the data is for foreign exchange
            cum: Whether to calculate the cumulative log returns
        """
        multiplier = 1
        if fx:
            multiplier = 10000
        if curPrc and prevPrc:
            return torch.log((curPrc/prevPrc))*multiplier
        else:
            if isinstance(prices, np.ndarray):
                prices = torch.from_numpy(prices)
                prices = torch.log(prices[1:]/prices[:-1])*multiplier
                return prices.numpy()
            else:
                return torch.log(prices[1:]/prices[:-1])*multiplier

    def inverse_targets(self, targets, transform = None):
        """
        Inverses the targets using the transform that was used to transform the data.

        Parameters:
            targets: The targets to inverse
            transform: The transform that was used to transform the data. Defaults to the default_transform
        """
        if transform is None:
            transform = self.default_transform

        # if self.train_data.shape[-1] > 1:
        #     if isinstance(targets, np.ndarray):
        #         targets = np.expand_dims(transform.inverse_transform(np.tile(targets, self.train_data.shape[-1]))[:,0], axis = 1)
        #     elif isinstance(targets, torch.Tensor):
        #         targets = transform.inverse_transform(targets.repeat(1, self.train_data.shape[-1]))[:,0].unsqueeze(1)
        # else:
        #     targets = transform.inverse_transform(targets)
        
        # transform the targets using the same transform that was used on the train_data repeat 
        if self.train_data.shape[-1] > 1:
            num_features = targets.shape[-1]
            if isinstance(targets, np.ndarray):
                targets_placeholder = np.zeros((targets.shape[0], 1))
                for i in range(self.train_data.shape[-1] - targets.shape[-1]):
                    targets = np.concatenate((targets, targets_placeholder), axis=1)
                targets = transform.inverse_transform(targets)[:, 0:(num_features)]
            elif isinstance(targets, torch.Tensor):
                targets_placeholder = torch.zeros((targets.shape[0], 1))
                for i in range(self.train_data.shape[-1]-targets.shape[-1]):
                    targets = torch.cat((targets, targets_placeholder), dim=1)
                targets = transform.inverse_transform(targets)[:,0:(num_features)]
        else:
            targets = transform.inverse_transform(targets)
        return targets

    def set_seed(self, seed):
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        np.random.seed(seed)




import torch
import torch.nn as nn
import torch.nn.functional as F

import logging

from lxfx.utils import createLogger

class FCBlock(nn.Module): # ForecastBlock
    def __init__(self, in_features, hidden_size, out_size, nature = "lstm", dropout= 0.2,
                 num_layers = 1, bidirectional = False, activation = "tanh",
                 use_batch_norm = False, pass_block_hidden_state = False):
        """
        Initializes the FCBlock with the provided parameters.

        Parameters:
            in_features (int): The number of input features.
            hidden_size (int): The size of the hidden layer.
            out_size (int): The size of the output layer.
            nature (str): The type of the block, one of "lstm", "rnn", "gru".
            dropout (float): The dropout rate.
            num_layers (int): The number of layers.
            bidirectional (bool): Whether the block is bidirectional.
            activation (str): The activation function, one of "tanh", "relu".
            use_batch_norm (bool): Whether to use batch normalization.
            pass_block_hidden_state: Whether to pass the hidden state of the first lstm to the next
        """
        super(FCBlock, self).__init__()
        self.nature = nature
        self.activation = activation
        if self.activation == "tanh":
            self.activation_function = nn.Tanh()
        elif self.activation == "relu":
            self.activation_function = nn.ReLU()
        self.num_layers = num_layers
        self.in_features = in_features
        self.hidden_size1 = hidden_size 
        self.output_size = out_size
        self.bidirectional = bidirectional
        self.pass_block_hidden_state = pass_block_hidden_state
        self.use_batch_norm = use_batch_norm
        self.dropout = dropout if self.num_layers > 1 else 0
        self.hidden_size2 = self.hidden_size1*2 if self.bidirectional is True else self.hidden_size1
        if self.nature == "lstm":
            self.l1 = nn.LSTM(self.in_features,hidden_size=self.hidden_size1, dropout=self.dropout,  num_layers=self.num_layers, bidirectional=self.bidirectional, batch_first=True)
            self.l2 = nn.LSTM(self.hidden_size2,hidden_size=self.output_size, dropout=self.dropout,  num_layers=self.num_layers, bidirectional=self.bidirectional, batch_first=True)
        elif self.nature == "rnn":
            self.l1 = nn.RNN(self.in_features,hidden_size=self.hidden_size1, dropout=self.dropout,  num_layers=self.num_layers, bidirectional=self.bidirectional, batch_first=True)
            self.l2 = nn.RNN(self.hidden_size2,hidden_size=self.output_size, dropout=self.dropout,  num_layers=self.num_layers, bidirectional=self.bidirectional, batch_first=True)            
        elif self.nature == "gru":
            self.l1 = nn.GRU(self.in_features,hidden_size=self.hidden_size1, dropout=self.dropout,  num_layers=self.num_layers, bidirectional=self.bidirectional, batch_first=True)
            self.l2 = nn.GRU(self.hidden_size2,hidden_size=self.output_size, dropout=self.dropout,  num_layers=self.num_layers, bidirectional=self.bidirectional, batch_first=True)

        # Add BatchNorm1d layers if use_batch_norm is True
        if self.use_batch_norm:
            bn_hidden_size2 = self.output_size*2 if self.bidirectional else self.output_size
            self.batch_norm1 = nn.BatchNorm1d(self.hidden_size2)
            self.batch_norm2 = nn.BatchNorm1d(bn_hidden_size2)

    def forward(self,x, prev_states = None):
        if self.nature == "lstm":
            output, (h1, c1) = self.l1(x, prev_states)
        else:
            output, h1 = self.l1(x, prev_states)

        # Apply BatchNorm1d if enabled
        # Note: batch_norm expects the input shape of (batch_size ,num_features, seq_length)
        # yet LSTMS, GRUS, RNNS, output shape is ( batch_size, seq_length, num_features) when batch_first is True
        if self.use_batch_norm:
            output = self.batch_norm1(output.transpose(1, 2)).transpose(1, 2)

        output = self.activation_function(output)

        # Pass to the next layer
        if self.nature == "lstm":
            if self.pass_block_hidden_state:
                output, (h2, c2) = self.l2(output, (h1, c1))
            else:
                output, (h2, c2) = self.l2(output)
            if self.use_batch_norm:
                output = self.batch_norm2(output.transpose(1, 2)).transpose(1, 2)
            return output, (h2, c2)
        else:
            if self.pass_block_hidden_state:
                output, h2 = self.l2(output, h1)
            else:
                output, h2 = self.l2(output)
            if self.use_batch_norm:
                output = self.batch_norm2(output.transpose(1, 2)).transpose(1, 2)
            return output, h2
        

class FxFCModel(nn.Module): # fxForeCastModel
    def __init__(self, num_features,
                 block_type = "lstm",
                 out_features = 1,
                 units:list = None,
                 num_layers = 1,
                 is_encoder = False,
                 encoder_latent_dim = None,
                 is_decoder = False, 
                 out_units:list = None,
                 activation:str = "tanh",
                 bidirectional = False, 
                 pass_states = False,
                 use_batch_norm = False, 
                 pass_block_hidden_state = False, 
                 decoder_out_features = None,
                 dropout = 0.2):
        """
        Parameters:
            num_features: the number of features per sequence eg if we have a seq [1., 2., 3., 4.] then the n_features = 4
            block_type: This is one of "lstm", "rnn", "gru" for the FCBlocks
            out_features: the number of features that we are trying to predict forexample if we trying to predict the close and moving average then this = 2
            units: A list of hidden sizes as they are supposed to propagate through the blocks
            is_encoder: Whether the model is to be used as an encoder in an autoencoder architecture
            encoder_latent_dim: the dimension for the latent representation of the encoder
            is_decoder: Whether the model is to be used as a decoder in an autoencoder Architecture
            out_units (list): The output sizes of the the blocks. This also affects the input shapes of the block of the preceeding blocks after the first has been set
            activation: The activation to use. One of "tanh", "relu"
            bidirectional: Whether the FCBlocks are bidirectional
            pass_states: Whether to pass the states to the next layer
            use_batch_norm: Whether to use batch normalization in the FCBlocks
            pass_block_hidden_state: Whether to pass the hidden state of the first lstm in the fcblock to the next
            decorder_out_features: if the model is to be used as a decoder then this is the actual number of features that we are trying to predict forexample if we trying to predict the close and moving average then this = 2
            dropout: The dropout rate
        todo:
            Initialize a list of bidirectional states and num_layers states for the FCBlocks(this requires not only saving 
            the model state dict but in order to reconstruct the model you must have saved these states in an extenal file)
        """
        super(FxFCModel, self).__init__() 

        self.model_type = "FxFCModel"

        self.num_features = num_features
        self.blocks = nn.ModuleList()
        self.block_type = block_type
        self.units = units
        self.out_units = out_units
        self.bidirectional = bidirectional
        self.num_layers = num_layers
        self.is_encoder = is_encoder
        self.use_batch_norm = use_batch_norm
        self.encoder_latent_dim = encoder_latent_dim
        self.decoder_out_features = decoder_out_features
        self.is_decoder = is_decoder
        self.activation = activation
        self.pass_states = pass_states
        self.dropout = dropout
        self.pass_block_hidden_state = pass_block_hidden_state
        self.console_logger = createLogger(logging.INFO, is_consoleLogger=True)

        # if self.out_units and len(self.out_units) == len(self.units):
        for i in range(len(self.units)):
            if i == 0:
                in_features = self.num_features   
            else:
                if not self.out_units:
                    in_features = self.units[i-1] if not self.bidirectional else self.units[i-1]*2
                else:
                    in_features = self.out_units[i-1] if not self.bidirectional else self.out_units[i-1]*2
            hidden_size = self.units[i]
            output_size = self.out_units[i] if self.out_units else hidden_size
            self.blocks.append(FCBlock(in_features,
                                       hidden_size,
                                       output_size,
                                       self.block_type,
                                       dropout=self.dropout,
                                       num_layers=self.num_layers,
                                       bidirectional=self.bidirectional,
                                       activation=self.activation,
                                       use_batch_norm = self.use_batch_norm,
                                       pass_block_hidden_state = self.pass_block_hidden_state))
        self.out_features = out_features
        if self.bidirectional:
            if  not self.out_units:
                self.fc_in_features = self.units[-1]*2
            else:
                self.fc_in_features = self.out_units[-1]*2
        else:
            if  not self.out_units:
                self.fc_in_features = self.units[-1]
            else:
                self.fc_in_features = self.out_units[-1]

        if self.is_encoder:
            out_features = self.encoder_latent_dim
        elif self.is_decoder:
            out_features = self.decoder_out_features
        
        self.fc = nn.Linear(in_features = self.fc_in_features, out_features=out_features)

    def config_dict(self):
        return {
            "num_features": self.num_features,
            "block_type": self.block_type,
            "units": self.units,
            "out_units": self.out_units,
            "bidirectional": self.bidirectional,
            "num_layers": self.num_layers,
            "is_encoder": self.is_encoder,
            "use_batch_norm": self.use_batch_norm,
            "encoder_latent_dim": self.encoder_latent_dim,
            "decoder_out_features": self.decoder_out_features,
            "is_decoder": self.is_decoder,
            "activation": self.activation,
            "pass_states": self.pass_states,
            "pass_block_hidden_state": self.pass_block_hidden_state,
            "dropout": self.dropout
        }

    def forward(self, x, prev_state = None):
        for idx, block in enumerate(self.blocks):
            # note
            # lstm output: o, (h, c)
            # rnn and gru output: o, h
            if idx > 0 and self.pass_states:
                if self.block_type == "lstm":
                    x, (h2, c2) = block(x, prev_state)
                    prev_state = (h2, c2)
                else:
                    x, h2 = block(x, prev_state)
                    prev_state = h2
            else:
                if self.block_type == "lstm":
                    x, (h2, c2) = block(x) # the 2 represents that these states are for the second block in the FCBlock
                    prev_state = (h2, c2)
                else:
                    x, h2 = block(x)    
                    prev_state = h2

        if self.is_encoder or self.is_decoder:
            return x, prev_state
        else:
            # h2 = h2.squeeze(0) # this only works on single single direction layers
            final_output = x[:, -1, :]
            x = self.fc(final_output)
            return x
            

class FxFCEncoder(nn.Module):
    def __init__(self, num_features,
                 block_type=None,
                 units=None,
                 out_units=None,
                 num_layers=1, 
                 activation="tanh",
                 latent_dim=None, 
                 dropout=0.2,
                 use_batch_norm=False,
                 bidirectional=False,
                 pass_block_hidden_state=False):
        """
        Parameters:
            num_features: the number of features per sequence
            block_type: This is one of "lstm", "rnn", "gru" for the FCBlocks
            units: A list of hidden sizes as they are supposed to propagate through the blocks
            out_units: The output sizes of the blocks. This also affects the input shapes of the block of the preceding blocks after the first has been set
            num_layers: The number of layers in the FCBlocks
            activation: The activation to use. One of "tanh", "relu"
            latent_dim: the dimension for the latent representation of the encoder
            dropout: The dropout rate
            use_batch_norm: Whether to use batch normalization in the FCBlocks
            bidirectional: Whether the FCBlocks are bidirectional
            pass_block_hidden_state: Whether to pass the hidden state of the first lstm in the fcblock to the next
        """
        super(FxFCEncoder, self).__init__()

        self.model_type = "FxFCEncoder"

        self.num_features = num_features 
        self.block_type = block_type 
        self.units = units 
        self.bidirectional = bidirectional
        self.out_units = out_units
        self.num_layers = num_layers 
        self.activation = activation
        self.latent_dim = latent_dim
        self.use_batch_norm = use_batch_norm
        self.dropout = dropout
        self.pass_block_hidden_state = pass_block_hidden_state

        self.encoder = FxFCModel(num_features=self.num_features,
                                 block_type=self.block_type,
                                 units=self.units,
                                 out_units=self.out_units,
                                 num_layers=self.num_layers,
                                 is_encoder=True,
                                 activation=self.activation,
                                 encoder_latent_dim=self.latent_dim,
                                 dropout=self.dropout,
                                 bidirectional=self.bidirectional, 
                                 use_batch_norm=self.use_batch_norm,
                                 pass_block_hidden_state=self.pass_block_hidden_state)

    def forward(self, input_data):
        # Initialize hidden and cell states
        h_t, c_t = (torch.zeros(self.num_layers, input_data.size(0), self.units[-1]).to(input_data.device),
                    torch.zeros(self.num_layers, input_data.size(0), self.units[-1]).to(input_data.device))
        
        # Initialize the encoded output
        input_encoded = torch.zeros(input_data.size(0), input_data.size(1), self.units[-1]).to(input_data.device)
        # Process each timestep in the sequence
        for t in range(input_data.size(1)):
            x_t = input_data[:, t, :].unsqueeze(1)  # Get the t-th timestep
            if self.block_type == "lstm":
                _, (h_t, c_t) = self.encoder(x_t, (h_t, c_t))
            else:
                _, h_t = self.encoder(x_t, h_t)
            input_encoded[:, t, :] = h_t[-1]  # Store the last layer's hidden state

        return self.encoder.fc(input_encoded[:, -1, :])

class FxFCDecoder(nn.Module):
    def __init__(self, latent_dim,
                 target_features=1,
                 block_type=None,
                 units=None,
                 out_units=None,
                 num_layers=1,
                 activation="tanh", 
                 dropout=0.2,
                 use_batch_norm=False,
                 bidirectional=False,
                 pass_block_hidden_state=False,
                 initialize_weights=False,
                 initializer_method=None,
                 future_pred_length=1):
        """
        Parameters:
            latent_dim: the number of features in the latent dim.
            target_features: the number of features that we are trying to predict
            block_type: This is one of "lstm", "rnn", "gru" for the FCBlocks
            units: A list of hidden sizes as they are supposed to propagate through the blocks
            out_units: The output sizes of the blocks. This also affects the input shapes of the block of the preceding blocks after the first has been set
            num_layers: The number of layers in the FCBlocks
            activation: The activation to use. One of "tanh", "relu"
            dropout: The dropout rate
            use_batch_norm: Whether to use batch normalization in the FCBlocks
            bidirectional: Whether the FCBlocks are bidirectional
            pass_block_hidden_state: Whether to pass the hidden state of the first lstm in the fcblock to the next
            initialize_weights: Whether to initialize the weights of the decoder
            initializer_method: The method to use to initialize the weights. One of "xavier", "uniform", "zeros"
            future_pred_length: The length of future predictions
        """
        super(FxFCDecoder, self).__init__()

        self.model_type = "FxFCDecoder"

        self.latent_dim = latent_dim
        self.block_type = block_type 
        self.units = units 
        self.out_units = out_units  
        self.num_layers = num_layers 
        self.activation = activation
        self.use_batch_norm = use_batch_norm
        self.target_features = target_features
        self.initialize_weights = initialize_weights
        self.initializer_method = initializer_method
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.pass_block_hidden_state = pass_block_hidden_state
        self.future_pred_length = future_pred_length

        self.decoder = FxFCModel(num_features=self.latent_dim,
                                 block_type=self.block_type, 
                                 units=self.units, 
                                 num_layers=self.num_layers, 
                                 is_decoder=True, 
                                 out_units=self.out_units, 
                                 activation=self.activation,
                                 dropout=self.dropout, 
                                 bidirectional=self.bidirectional,
                                 use_batch_norm=self.use_batch_norm, 
                                 decoder_out_features=self.target_features,
                                 pass_block_hidden_state=self.pass_block_hidden_state)

        self.tensor_initializer = TensorInitializer(method="xavier", num_layers=self.num_layers)

    def forward(self, encoded_input):
        outputs = []
        decoder_input = encoded_input
        # initialize the hidden states 
        # h_t, c_t = (self.tensor_initializer())
        # Note: encoded_input.size(0) is the batch_size
        prev_state = None
        if self.initialize_weights:
            if self.initializer_method == "xavier":
                h_t, c_t = (self.tensor_initializer.initializeTensor(),
                            self.tensor_initializer.initializeTensor())
                prev_state = (h_t, c_t)
            elif self.initializer_method == "zeros":
                h_t, c_t = (torch.zeros(self.num_layers, encoded_input.size(0), self.units[-1]).to(encoded_input.device),
                            torch.zeros(self.num_layers, encoded_input.size(0), self.units[-1]).to(encoded_input.device))
                prev_state = (h_t, c_t)
            else:
                prev_state = None
        else:
            prev_state = None

        # The encoder returns a tensor of size (batch_size, latent_dim)
        # Note: self.decoder expects a tensor of size (batch_size, seq_len, latent_dim)
        # Add a sequence length dimension to the encoded_input
        decoder_input = decoder_input.unsqueeze(1)

        for t in range(self.future_pred_length):
            if self.block_type == "lstm":
                if prev_state is not None:
                    output, (h_t, c_t) = self.decoder(decoder_input, prev_state)
                else:
                    output, (h_t, c_t) = self.decoder(decoder_input)
            else:
                if prev_state is not None:
                    output, h_t = self.decoder(decoder_input, prev_state)
                else:
                    output, h_t = self.decoder(decoder_input)

            decoder_input = output
            
            pred = self.decoder.fc(output[:, -1, :])
            pred = pred.unsqueeze(0)
            outputs.append(pred)

        outputs = torch.cat(outputs, dim=0) # size = (future_pred_length, batch_size, target_features)
        outputs_shape = outputs.shape
        future_dim, batch_dim, features_dim = outputs_shape[0], outputs_shape[1], outputs_shape[2]
        outputs = outputs.view((batch_dim, future_dim, features_dim)) # size = (batch_size, future_pred_length, target_features)
        return outputs

class FxFCAutoEncoder(nn.Module):
    def __init__(self, num_features,
                 target_features=1,
                 future_pred_length=1,
                 block_types=("lstm", "lstm"),
                 units=None,
                 out_units=None,
                 num_layers=(1, 1),
                 activations=("tanh", "tanh"),
                 latent_dim=32,
                 dropout=(0.2, 0.2),
                 bidirectional=(False, False),
                 use_batch_norm=(False, False),
                 pass_block_hidden_state=(False, False)):
        """
        Parameters:
            num_features: the number of features per sequence
            target_features: the number of features to predict
            future_pred_length: the length of future predictions
            block_types: A tuple of the block types for the encoder and decoder
            units: A tuple of the hidden sizes for the encoder and decoder
            out_units: A tuple of the output sizes for the encoder and decoder
            num_layers: A tuple of the number of layers for the encoder and decoder
            activations: A tuple of the activation functions for the encoder and decoder
            latent_dim: the dimension for the latent representation of the encoder
            dropout: A tuple of dropout rates for the encoder and decoder
            bidirectional: A tuple indicating if the encoder and decoder are bidirectional
            use_batch_norm: A tuple indicating if batch normalization is used in the encoder and decoder
            pass_block_hidden_state: A tuple indicating if the hidden state is passed in the encoder and decoder
        """
        super(FxFCAutoEncoder, self).__init__()

        self.model_type = "FxFCAutoEncoder"

        self.num_features = num_features
        self.target_features = target_features
        self.future_pred_length = future_pred_length
        self.block_types = block_types
        self.encoder_block_type = self.block_types[0]
        self.decoder_block_type = self.block_types[1]

        self.units = units
        self.encoder_units, self.decoder_units = self.units

        self.out_units = out_units
        if self.out_units:
            self.encoder_out_units = self.out_units[0]
            self.decoder_out_units = self.out_units[1]
        else:
            self.encoder_out_units = None
            self.decoder_out_units = None

        self.num_layers = num_layers
        self.encoder_num_layers = self.num_layers[0]
        self.decoder_num_layers = self.num_layers[1]

        self.activations = activations
        self.encoder_activation = self.activations[0]
        self.decoder_activation = self.activations[1]

        self.encoder_latent_dim = latent_dim

        self.dropout = dropout
        self.encoder_dropout = self.dropout[0]
        self.decoder_dropout = self.dropout[1]

        self.bidirectional = bidirectional
        self.encoder_bidirectional = self.bidirectional[0]
        self.decoder_bidirectional = self.bidirectional[1]

        self.use_batch_norm = use_batch_norm
        self.encoder_use_batch_norm = self.use_batch_norm[0]
        self.decoder_use_batch_norm = self.use_batch_norm[1]

        self.pass_block_hidden_state = pass_block_hidden_state
        self.encoder_pass_block_hidden_state = self.pass_block_hidden_state[0]
        self.decoder_pass_block_hidden_state = self.pass_block_hidden_state[1]

        self.encoder = FxFCEncoder(num_features=self.num_features, 
                                   block_type=self.encoder_block_type,
                                   units=self.encoder_units, 
                                   out_units=self.encoder_out_units,
                                   num_layers=self.encoder_num_layers, 
                                   activation=self.encoder_activation,
                                   latent_dim=self.encoder_latent_dim, 
                                   dropout=self.encoder_dropout,
                                   bidirectional=self.encoder_bidirectional, 
                                   use_batch_norm=self.encoder_use_batch_norm,
                                   pass_block_hidden_state=self.encoder_pass_block_hidden_state)

        self.decoder = FxFCDecoder(latent_dim=self.encoder_latent_dim,
                                   block_type=self.decoder_block_type,
                                   units=self.decoder_units,
                                   out_units=self.decoder_out_units,
                                   num_layers=self.decoder_num_layers,
                                   activation=self.decoder_activation,
                                   target_features=self.target_features,
                                   dropout=self.decoder_dropout,
                                   bidirectional=self.decoder_bidirectional,
                                   use_batch_norm=self.decoder_use_batch_norm,
                                   pass_block_hidden_state=self.decoder_pass_block_hidden_state,
                                   future_pred_length=self.future_pred_length)

    def config_dict(self):
        return {
            "num_features": self.num_features,
            "target_features": self.target_features,
            "future_pred_length": self.future_pred_length,
            "block_types": self.block_types,
            "units": self.units,
            "out_units": self.out_units,
            "num_layers": self.num_layers,
            "activations": self.activations,
            "latent_dim": self.encoder_latent_dim,
            "dropout": self.dropout,
            "bidirectional": self.bidirectional,
            "use_batch_norm": self.use_batch_norm,
            "pass_block_hidden_state": self.pass_block_hidden_state
        }

    def forward(self, input_seq):
        encoded_vector = self.encoder(input_seq)  # size = (batch_size, latent_dim)
        preds = self.decoder(encoded_vector)
        return preds

class TensorInitializer:
    def __init__(self, method = None, x: torch.Tensor = None, num_layers = 1,
                 hidden_size = None):
        self.method = method
        self.tensor = x
        self.num_layers = num_layers
        self.hidden_size = hidden_size

    def initializeTensor(self):
        self.p_h = torch.zeros(self.num_layers, self.tensor.size(0), self.hidden_size)
        if self.method == "xavier":
            return nn.init.xavier_normal_(self.p_h)
        else:
            return self.p_h


import logging
import random
import sys

def handleExit():
    """
    Handles exit scenarios for both Jupyter notebooks and normal Python scripts.
    """
    try:
        # Check if the environment is a Jupyter notebook
        get_ipython
        # Use IPython's exit function for notebooks
        from IPython.display import display, Javascript
        display(Javascript('IPython.notebook.kernel.execute("exit()")'))
        # Raise an exception to stop the current cell execution
        raise SystemExit("Exiting the notebook cell.")
    except NameError:
        import sys
        sys.exit()

def createLogger(log_level = None, name = None, mode = 'a', log_format = None, is_consoleLogger = False, filename = None):
    """
    Creates a logger object with the specified parameters.
    Parameters:
        log_level: The level of logging to be used. Defaults to logging.INFO
        name: The name of the logger. Defaults to a random number between 1 and 100
        mode: The mode of the file handler. Defaults to 'a'
        log_format: The format of the log message. Defaults to "%(asctime)s -%(levelname)s - %(filename)s:%(lineno)d - %(message)s"
        is_consoleLogger: Whether to log to the console. Defaults to False
        filename: The filename to be used for the file handler. Defaults to None
    """
    if log_level is None:
        log_level = logging.INFO
    if log_format is None:
        log_format = "%(asctime)s -%(levelname)s - %(filename)s:%(lineno)d - %(message)s"
    if is_consoleLogger is False:
        if name is None:
            name = "logger"+str(random.randint(1,100))
    else:
        if name is None:
            name = __name__
    logger = logging.getLogger(name)
    if is_consoleLogger is True:
        file_handler = logging.StreamHandler()
    else:
        if filename is None:
            filename = sys.executable.rstrip(".py")+".log"
        file_handler = logging.FileHandler(filename=filename, mode=mode)
    file_handler.setLevel(log_level)
    formater = logging.Formatter(log_format)
    file_handler.setFormatter(formater)
    logger.addHandler(file_handler)
    logger.setLevel(log_level)
    return logger


import logging
import json
import os
import atexit

from lxfx.utils import createLogger, handleExit
from lxfx.data import TimeSeriesDatasetManager
from lxfx.trainers import ModelTrainer
from lxfx.predPlotter import PredPlotter
from lxfx.models import FxFCModel, FxFCAutoEncoder

class XFX:
    def __init__(self, config_file_path = None, _config_dict = None, 
                 model_type = None, manual_seed = None):
        self.console_logger = createLogger(is_consoleLogger=True)
        self.config_file_path = config_file_path
        self._config_dict = _config_dict
        self.model_type = model_type
        self.dataset_manager = None 

        self.dataset_config = None
        self.trainer_config = None
        self.FxFCModel_config = None
        self.FxFCAutoEncoder_config = None

        self.trainer = None 
        self.dataset_manager = None
        self.model = None

        self.get_config()

        self.dataset_manager = self.get_dataset_manager()
        if manual_seed:
            self.dataset_manager.set_seed(manual_seed)
        self.model = self.get_model()
        self.trainer = self.get_trainer()

        atexit.register(self.save_config)

    def get_config(self):
        if self.config_file_path:
            try:
                self.config = json.load(open(self.config_file_path))
            except Exception as e:
                self.console_logger.error(f"Error loading config file: {e}")
        else:
            self.config = self._config_dict

        # get datasetManager and trainer 
        try:
            self.dataset_config = self.config["TimeSeriesDatasetManager"]
        except Exception as e:
            self.console_logger.error(f"Error loading dataset config: {e}")
        try:
            self.trainer_config = self.config["ModelTrainer"]
        except Exception as e:
            self.console_logger.error(f"Error loading trainer config: {e}")

        # FxFCModel config
        try:
            self.FxFCModel_config = self.config["FxFCModel"]
        except Exception as e:
            if self.model_type == "FxFCModel":
                self.console_logger.error(f"Error loading FxFCModel config: {e}")
                self.console_logger.error("Please check the config file for the FxFCModel section.")
            else:
                self.FxFCModel_config = None

        try:
            self.FxFCAutoEncoder_config = self.config["FxFCAutoEncoder"]
        except Exception as e:
            if self.model_type == "FxFCAutoEncoder":
                self.console_logger.error(f"Error loading FxFCAutoEncoder config: {e}")
                self.console_logger.error("Please check the config file for the FxFCAutoEncoder section.")
            else:
                self.FxFCAutoEncoder_config = None

    def get_dataset_manager(self):
        if self.dataset_manager is None:
            self.dataset_manager = TimeSeriesDatasetManager(
                data=None,
                file_path=self.dataset_config.get("file_path"),
                targets=None,
                future_pred_length=self.dataset_config.get("future_pred_length", 1),
                targetIndices=self.dataset_config.get("targetIndices", None),
                target_names=self.dataset_config.get("target_names"),
                drop_names=self.dataset_config.get("drop_names", None),
                date_col=self.dataset_config.get("date_col"),
                sequence_length=self.dataset_config.get("sequence_length", 5),
                stride=self.dataset_config.get("stride", 1),
                scaler_feature_range=self.dataset_config.get("scaler_feature_range", [0, 1]),
                transform=self.dataset_config.get("transform", None),
                use_default_scaler=self.dataset_config.get("use_default_scaler", True),
                split_criterion=self.dataset_config.get("split_criterion", [0.7, 0.15, 0.15]),
                batchsize=self.dataset_config.get("batchsize", 8),
                is_fx=self.dataset_config.get("is_fx", False),
                transform_targets=self.dataset_config.get("transform_targets", True),
                index_col=self.dataset_config.get("index_col", None),
                manual_seed=self.dataset_config.get("manual_seed", None),
                is_autoregressive=self.dataset_config.get("is_autoregressive", False),
                test_sequence_length=self.dataset_config.get("test_sequence_length", None), 
                test_stride=self.dataset_config.get("test_stride", None)
            )
        return self.dataset_manager

    def get_trainer(self):
        if self.trainer is None:
            self.trainer = ModelTrainer(
                model=self.model,
                dataset_manager=self.dataset_manager,
                loss_fn=self.trainer_config.get("loss_fn", "mse"),
                optimizer=self.trainer_config.get("optimizer", "adam"),
                epochs=self.trainer_config.get("epochs", 3),
                accuracy_fn=self.trainer_config.get("accuracy_fn", None),
                lr=self.trainer_config.get("lr", 0.001),
                project_path=self.trainer_config.get("project_path"),
                is_logging=self.trainer_config.get("is_logging", True),
                resume=self.trainer_config.get("resume", False),
            )
        return self.trainer

    def get_model(self): 
        if self.model is None:
            if self.model_type == "FxFCModel":
                self.model = FxFCModel(
                    num_features=self.FxFCModel_config.get("num_features"),
                    block_type=self.FxFCModel_config.get("block_type", "lstm"),
                    out_features=self.FxFCModel_config.get("out_features", 1),
                    units=self.FxFCModel_config.get("units"),
                    num_layers=self.FxFCModel_config.get("num_layers", 1),
                    is_encoder=self.FxFCModel_config.get("is_encoder", False),
                    encoder_latent_dim=self.FxFCModel_config.get("encoder_latent_dim"),
                    is_decoder=self.FxFCModel_config.get("is_decoder", False),
                    out_units=self.FxFCModel_config.get("out_units"),
                    activation=self.FxFCModel_config.get("activation", "tanh"),
                    bidirectional=self.FxFCModel_config.get("bidirectional", False),
                    pass_states=self.FxFCModel_config.get("pass_states", False),
                    use_batch_norm=self.FxFCModel_config.get("use_batch_norm", False),
                    pass_block_hidden_state=self.FxFCModel_config.get("pass_block_hidden_state", False),
                    decoder_out_features=self.FxFCModel_config.get("decoder_out_features")
                )
            elif self.model_type == "FxFCAutoEncoder":
                self.model = FxFCAutoEncoder(
                    num_features=self.FxFCAutoEncoder_config.get("num_features"),
                    target_features=self.FxFCAutoEncoder_config.get("target_features", 1),
                    future_pred_length=self.FxFCAutoEncoder_config.get("future_pred_length", 1),
                    block_types=self.FxFCAutoEncoder_config.get("block_types", ["lstm", "lstm"]),
                    units=self.FxFCAutoEncoder_config.get("units"),
                    out_units=self.FxFCAutoEncoder_config.get("out_units"),
                    num_layers=self.FxFCAutoEncoder_config.get("num_layers", [1, 1]),
                    activations=self.FxFCAutoEncoder_config.get("activations", ["tanh", "tanh"]),
                    latent_dim=self.FxFCAutoEncoder_config.get("latent_dim", 32),
                    dropout=self.FxFCAutoEncoder_config.get("dropout", [0.2, 0.2]),
                    bidirectional=self.FxFCAutoEncoder_config.get("bidirectional", [False, False]),
                    use_batch_norm=self.FxFCAutoEncoder_config.get("use_batch_norm", [False, False]),
                    pass_block_hidden_state=self.FxFCAutoEncoder_config.get("pass_block_hidden_state", [False, False])
                )
            else:
                self.console_logger.error(f"Invalid model type: {self.model_type}")
                handleExit()
        return self.model
        
    def train(self, epochs = None):
        self.trainer.trainModel(epochs=epochs)
        self.save_config()

    def predict(self):
        self.trainer.predict()

    def save_config(self):
        config_dict = {}
        config_dict["model_type"] = self.model_type
        config_dict["model"] = self.model.config_dict()
        config_dict["dataset"] = self.dataset_manager.config_dict()
        config_dict["trainer"] = self.trainer.config_dict()

        project_dir = self.trainer.project_path
        config_file_path = os.path.join(project_dir, "config.json")
        with open(config_file_path, "w") as f:
            json.dump(config_dict, f, indent=4)
        print(f"Config file saved to: {config_file_path}")

    def plotPredictions(self, preds, figsize = (6,6), test_only=False,
                        _len = 100, test_plots = None):
        """
        Plots a graph(s) showing the predictions against the actual values.
        """
        predPlotter = PredPlotter(preds, self.dataset_manager, test_only=test_only,
                                  length = _len, figsize = figsize, test_plots = test_plots)
        predPlotter.plotPredictions()


import torch
import torch.nn as nn
import numpy as np

class RMSELoss(nn.Module):
    def __init__(self, eps = 1e-6):
        super(RMSELoss, self).__init__()
        self.mse = nn.MSELoss()
        self.eps = eps

    def forward(self, pred, true):
        loss = torch.sqrt(self.mse(pred, true)+self.eps)
        return loss

class TorchMinMaxScaler:
    def __init__(self, feature_range=(0, 1)):
        self.min = None
        self.max = None
        self.scale = None
        self.min_range, self.max_range = feature_range

    def fit(self, data):
        # Ensure data is a 2D PyTorch tensor
        if isinstance(data, np.ndarray):
            data = torch.tensor(data, dtype=torch.float32)
        if data.dim() == 1:
            data = data.unsqueeze(1)
        
        # Calculate the min and max for each feature
        self.min = data.min(dim=0, keepdim=True).values
        self.max = data.max(dim=0, keepdim=True).values
        self.scale = (self.max_range - self.min_range) / (self.max - self.min)

    def transform(self, data):
        return (data - self.min) * self.scale + self.min_range

    def inverse_transform(self, data):
        return (data - self.min_range) / self.scale + self.min

class TorchStandardScaler:
    def __init__(self):
        self.mean = None 
        self.std = None 

    def fit(self, data):
        self.mean = data.mean(dim = 0, keepdim = True)
        self.std = data.std(dim = 0, keepdim = True, unbiased = False)

    def transform(self, data):
        return (data-self.mean)/self.std
    
    def inverse_transform(self, data):
        return data*self.std + self.mean

def MAPE(y_pred, y_true):
    """This function implements the mean absolute percentage error accuracy where we are just
    trying to see how well the predicted value corresponds to the true value.
    Parameters:
        y_pred: predicted value 
        y_true: true value 
    Returns:
        a float representing the percentage accuracy
    """
    return 1-torch.mean(torch.abs((y_true - y_pred)/y_true))


import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
import os
from torch.utils.data import DataLoader
import tqdm
import logging

from lxfx.data import TimeSeriesDatasetManager
from lxfx.utils import createLogger
from lxfx.metrics import RMSELoss

class ModelTrainer():
    def __init__(self, 
                model = None,
                dataset_manager:TimeSeriesDatasetManager = None,
                train_dataloader:DataLoader = None,
                loss_fn = None,
                optimizer = None,
                epochs = 3,
                validation_dataloader:DataLoader = None,
                test_dataloader = None,
                accuracy_fn = None,
                lr= 0.001,
                project_path = None, 
                is_logging = True,
                resume = False,
                targets_transformed = False):
        """Class to handle training of the model on a train data from the data_dataloader and validation data from
        the validation_dataloader.
        Parameters:
            model: model to undergo training
            dataset_manager: The datasetManager used to generate the dataloaders
            train_dataloader: The dataloader that is supposed to load the training feature and labels
            loss_fn: Loss function to be used. e.g loss_fn = MSELoss(), fit(loss_fn = loss_fn), This can also be a function
            optmizer: The optmizimzer to be used to edit  the model parameters
            epochs: Number of times to train the model on the data (num of different times the model has to see the data)
            validation_dataloader: Dataloader that is to load the validation data.
            test_dataloader: Dataloader thtat is to load the test data for prediction.
            accuracy_fn: The accuracy function or class to be used.
            lr: The learning rate to be used by the optimizer. Defaults to 0.001
            project_path: This is the path to the directory where the best model is to be saved, results and so on as the model is training
            is_logging: whether to log training and other kinds of information into a file in the project dir path
            resume: Whether to resume training from the last saved model.
            targets_transformed: Wether or not the targets were transformed
        Examples:
            >>> train_dataloader = DataLoader(train_dataset, shuffle = False, batch_size = 32)
            >>> validation_dataloader = DataLoader(val_dataset, shuffle = False, batch_size = 32)
            >>> test_dataloader = DataLoader(test_dataset, shuffle = False, batch_size= len(test_dataset))
            >>> loss_fn = "mse" or MSELoss()
            >>> optimizer = "adam" or Adam(model.paramters(), lr = 0.001)
        """
        self.console_logger = createLogger(logging.INFO, is_consoleLogger=True)

        self.model = model
        self.loaded_model = None
        self.targets_transformed= targets_transformed
        self.dataset_manager = dataset_manager 

        if self.dataset_manager:
            self.train_dataloader, self.validation_dataloader, self.test_dataloader = self.dataset_manager.dataloaders()
            self.targets_transformed = self.dataset_manager.transform_targets
        else:
            self.train_dataloader = train_dataloader
            self.validation_dataloader = validation_dataloader
            self.test_dataloader = test_dataloader

        self.loss_fn = loss_fn
        self.getloss_fn()

        self.optmizer = optimizer
        self.optim_lr = lr
        self.getOptimizer()

        self.acc_fn = accuracy_fn
        self.epochs = epochs

        self.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        self.model.to(self.device)
        self.stat_logs = {}
        self.resume = resume

        self.min_loss_ = float('inf')
        self.project_path = project_path
        self.project_path_exists = False

        if self.project_path_exists:
            self.resume = True

        if self.project_path:
            self.project_path = self.getLoggingDir()
        self.is_logging = is_logging
        if not self.project_path:
            self.is_logging = False
        self.bestModelPath = os.path.join(self.project_path, "best.pth") if self.project_path else None
        self.lastModelPath = os.path.join(self.project_path, "last.pth") if self.project_path else None

        if not self.resume:
            if not self.project_path_exists:
                if self.bestModelPath:
                    if os.path.exists(self.bestModelPath):
                        os.remove(self.bestModelPath)
                if os.path.exists(self.lastModelPath):
                    os.remove(self.lastModelPath)
            else:
                self.model = self.loadModel()
        else:
            self.model = self.loadModel

        self.training_logs = {
            "stat_logs":{},
            "min_loss_":self.min_loss_,
            "optmizer":None,
            "loss_fn":None
        }
        self.early_stopping_counte = 0
        self.scheduler = optim.lr_scheduler.ReduceLROnPlateau(self.optmizer, "min", patience=2, verbose=True)# change the learning rage based on the validation loss i.e if the val loss does not improve for 2 epochs then reduce the learning rate by half

        if self.project_path:
            self.trainLogger_filepath= os.path.join(self.project_path, "training_logs.txt")
            self.training_file_logger = createLogger(log_level=logging.INFO,
                                                    name="trainingLogger",
                                                    filename= self.trainLogger_filepath)

        self.model_type = self.model.model_type

    def loadModel(self):
        print(f"Loading model from path: {self.bestModelPath}")
        loaded_model = torch.load(self.bestModelPath)
        loaded_model.to(self.device)
        return loaded_model

    def config_dict(self):
        return {
            "model": str(self.model),
            "dataset_manager": str(self.dataset_manager),
            "epochs": self.epochs,
            "learning_rate": self.optim_lr,
            "loss_function": str(self.loss_fn),
            "optimizer": str(self.optmizer),
            "device": str(self.device),
            "project_path": self.project_path,
            "is_logging": self.is_logging,
            "resume": self.resume,
            "targets_transformed": self.targets_transformed
        }

    def getLoggingDir(self):
        # Get the directory where to log all the obtained information including models, weights, results, images and so on 
        if "Train_" in self.project_path:
            self.project_path_exists = True
            print(f"XFX project found in dirirectory: {self.project_path}\nUsing project directory.")
            return self.project_path
        
        if self.project_path:
            available_dirs = os.listdir(self.project_path)
            train_dir_count = 0
            for dir_name in available_dirs:
                if os.path.isdir(os.path.join(self.project_path, dir_name)) and dir_name.startswith("Train_"):
                    try:
                        count = int(dir_name.split("_")[-1])
                        if count > train_dir_count:
                            train_dir_count = count
                    except ValueError:
                        continue  # Skip directories that don't end with a number
            train_dir_name = f"Train_{train_dir_count + 1}"
            logging_dir = os.path.join(self.project_path, train_dir_name)
            try:
                os.mkdir(logging_dir)
                return logging_dir
            except PermissionError:
                self.console_logger.warning(f"The program does not have enough permissions to create a directory in path {self.project_path}. Consider changing permissions.")
            except Exception as e:
                self.console_logger.error(f"Encountered error {e}")
                return self.project_path

    def getloss_fn(self):
        if self.loss_fn == "mse":
            self.loss_fn = nn.MSELoss()
        elif self.loss_fn == "rmse":
            self.loss_fn = RMSELoss()

    def getOptimizer(self):
        if self.optmizer == "adam":
            self.optmizer = optim.Adam(params = self.model.parameters(), lr=self.optim_lr)

    def resume(self, ):
        """
        Resumes the training of the model from the last saved model.
        """
        pass 

    def computeArModelLoss(self, preds:torch.Tensor, labels:torch.Tensor):
        preds = preds.unsqueeze(1)
        return self.loss_fn(preds, labels)

    def computeLoss(self, features, labels, train = False):
        features, labels = features.to(self.device), labels.to(self.device)
        preds = self.model(features)

        # transform the preds if the targets weren't transformed
        if not self.targets_transformed:
            preds = self.dataset_manager.inverse_targets(preds)

        loss = self.computeArModelLoss(preds, labels)

        if train:
            loss.backward()
            nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=1.0) # gradient clipping
        if isinstance(loss, torch.Tensor):
            loss = loss.item()
        if self.acc_fn:
            acc = self.acc_fn(preds.cpu(), labels.cpu())
            if isinstance(acc, torch.Tensor):
                acc = acc.item()
            return loss, acc
        return loss, None

    def evaluateModel(self, testDataLoader:DataLoader = None):
        """
        Evaluates the model on the test data.

        Parameters:
            testDataLoader (DataLoader): The dataloader that is supposed to load the test data.

        Returns:
            avg_loss (float): The average loss of the model on the test data.
            avg_acc (float): The average accuracy of the model on the test data.
        """

        if not testDataLoader:
            if self.dataset_manager:
                testDataLoader =self.test_dataloader

        self.model.eval()
        with torch.inference_mode():
            total_loss = 0
            total_acc = 0
            for idx, (features, labels) in enumerate(testDataLoader):
                loss, acc = self.computeLoss(features, labels)
                total_loss += loss
                if self.acc_fn:
                    total_acc += acc
            avg_loss = total_loss / len(testDataLoader)
            if self.acc_fn:
                avg_acc = total_acc / len(testDataLoader)
                print(f"Test Loss: {avg_loss:.5f}, Test Acc: {avg_acc:.5f}")
            else:
                avg_acc = None
                print(f"Test Loss: {avg_loss:.5f}")
        return avg_loss, avg_acc

    def validateModel(self):
        self.model.eval()
        with torch.inference_mode():
            total_loss = 0
            total_acc = 0
            for idx, (features, labels) in enumerate(self.validation_dataloader):
                loss, acc = self.computeLoss(features, labels)
                total_loss += loss
                if self.acc_fn:
                    total_acc += acc
            avg_loss = total_loss / len(self.validation_dataloader)
            avg_acc = total_acc / len(self.validation_dataloader) if self.acc_fn else None
        return avg_loss, avg_acc

    def trainModel(self, epochs = None, verbose_level = "min"):
        """
        Trains the model on the train data.

        Parameters:
            epochs (int): The number of epochs to train the model.
            verbose (bool): Whether to show alot of training steps per epoch
            verbose_level( one if min, mid, max): how much of logging inf to show on the screen
        Returns:
            train_loss_history (list): The history of the training loss.
            train_acc_history (list): The history of the training accuracy.
            val_loss_history (list): The history of the validation loss.
            val_acc_history (list): The history of the validation accuracy.
        """
        train_loss_history = []
        train_acc_history = []
        val_loss_history = []
        val_acc_history = []
        if verbose_level == "max":
            display_steps = len(self.train_dataloader)//10 
        elif verbose_level == "mid":
            display_steps = len(self.train_dataloader)//0
        else: # min falls under here
            display_steps = None

        if epochs:
            self.epochs= epochs

        for epoch in tqdm.tqdm(range(0, self.epochs)):
            print(f"Epoch: {epoch}")
            self.model.train()
            epoch_total_loss = 0
            epoch_total_acc = 0
            for idx, (features, labels) in enumerate(self.train_dataloader):
                self.optmizer.zero_grad()
                train_loss, train_acc = self.computeLoss(features, labels, train = True)
                epoch_total_loss += train_loss
                if self.acc_fn:
                    epoch_total_acc += train_acc
                self.optmizer.step()

                if self.acc_fn:
                    step_log_info = f"\tStep: {idx}/{len(self.train_dataloader)} Train Loss: {train_loss:.5f}, Train Acc: {train_acc:.5f}"
                    if display_steps:
                        if idx % display_steps == 0:
                            print(step_log_info)
                else:
                    step_log_info = f"\tStep: {idx}/{len(self.train_dataloader)} Train Loss: {train_loss:.5f}"
                    if display_steps:
                        if idx % display_steps == 0:
                                print(step_log_info)

                if self.project_path and self.is_logging:
                    self.training_file_logger.info(step_log_info)

            val_loss, val_acc = self.validateModel()

            train_loss = epoch_total_loss / len(self.train_dataloader)
            train_acc = epoch_total_acc / len(self.train_dataloader) if self.acc_fn else None

            train_loss_history.append(train_loss)
            val_loss_history.append(val_loss)

            self.scheduler.step(val_loss) # learning rate scheduling

            if self.acc_fn:
                train_acc_history.append(train_acc)
                val_acc_history.append(val_acc)
                epoch_log_info = f"Epoch: {epoch+1}, Train Loss: {train_loss:.5f}, Train Acc: {train_acc:5f}, Val Loss: {val_loss:.5f}, Val Acc: {val_acc:.5f}"
                print(epoch_log_info)
            else:
                epoch_log_info = f"Epoch: {epoch+1}, Train Loss: {train_loss:.5f}, Val Loss: {val_loss:.5f}"
                print(epoch_log_info)

            if self.project_path and self.is_logging:
                self.training_file_logger.info(epoch_log_info)  

            if val_loss < self.min_loss_:
                print(f"Model val_loss improved from {self.min_loss_} to {val_loss}")
                self.min_loss_ = val_loss
                if self.project_path:
                    torch.save(self.model, self.bestModelPath)
                    print(f"Model Saved to file: {self.bestModelPath}")
            else:
                print(f"Model val_loss did not improve from {self.min_loss_}")
        self.stat_logs["train_loss_history"] = train_loss_history
        self.stat_logs["train_acc_history"] = train_acc_history 
        self.stat_logs["val_loss_history"] = val_loss_history 
        self.stat_logs["val_acc_history"] = val_acc_history
        return train_loss_history, train_acc_history, val_loss_history, val_acc_history

    def visualizeResults(self, dirpath = None, show= True, format = None, figsize = (15, 20)):
        """ 
        visualize the results from the training and or save them
        Parameters:
            dirpath: The directory to which to save the image files
            show: Whether or not to show the images.
            format: file format to which to save the files. Default is png
            figsize: The figure size to use when plotting eg figsize = (15, 20) #<== default
        """
        fig1 = plt.figure(figsize = figsize)

        # loss vs epochs 
        if self.stat_logs:
            axs1 =fig1.add_subplot(1,1,1)
            axs1.plot(range(self.epochs), self.stat_logs["train_loss_history"], label = "train_loss")
            if len(self.stat_logs["val_loss_history"]) != 0:
                axs1.plot(range(self.epochs), self.stat_logs["val_loss_history"], label = "val_loss")
            fig1.legend()
            if dirpath:
                fname = os.path.join(dirpath, "train_and_val_losses_vs_epochs.png")
                fig1.savefig(fname = fname)
            elif self.project_path:
                fname = os.path.join(self.project_path, "train_and_val_losses_vs_epochs.png")
                fig1.savefig(fname = fname)

            # accuracies vs epochs 
            if self.acc_fn:
                fig2 = plt.figure(figsize = figsize)
                axs2 = fig2.add_subplot(1,1,1)
                axs2.plot(range(self.epochs), self.stat_logs["train_acc_history"], label="train_accuracy")
                if len(self.stat_logs["val_loss_history"]) != 0:
                    axs2.plot(range(self.epochs), self.stat_logs["val_acc_history"], label= "val_accuracy")
                fig2.legend()
            if dirpath:
                fname = os.path.join(dirpath, "train_and_val_accuracies_vs_epochs.png")
                fig1.savefig(fname = fname)
            elif self.project_path:
                fname = os.path.join(self.project_path, "train_and_val_accuracies_vs_epochs.png")
                fig1.savefig(fname = fname)

    def predict(self, saved = True):
        """Make prediction from the test data."""
        if saved:
            if not self.project_path_exists:
                if self.project_path:
                    print(f"Loading model from path: {self.bestModelPath}")
                    self.loaded_model = torch.load(self.bestModelPath)
                else:
                    print("Project directory not specified")
            else:
                self.loaded_model = self.model
        else:
            self.loaded_model = self.model
		
        self.loaded_model.to(self.device)

        if len(self.test_dataloader) != 0:
            self.loaded_model.eval()
            outputs = []

        # Note: When the model is auto
        with torch.inference_mode():
            for idx, (slice_, __) in enumerate(self.test_dataloader):
                slice_ = slice_.to(self.device)
                output = self.loaded_model(slice_)
                if self.model_type == "FxFCAutoEncoder": # This predicts future values autoregressively
                    outputs.append(output.cpu()) # remove the batch dimension
                else:
                    if output.shape[-1] == 1:
                        outputs.append([output.cpu().item()])
                    else:
                        outputs.append(output.cpu().tolist())
        if self.model_type == "FxFCAutoEncoder":
            return outputs # This is a list of tensors of the different future predictions for each batch in the test_dataloader
        return torch.tensor(outputs)
